#include <mp4lib/mp4.h>
#include <mp4lib/genres.h>

/* This code more or less wholesale lifted from the xmms Vorbis plugin */
#ifdef HAVE_ICONV
#include <iconv.h>

char* convert_string(const char *string)
{
  size_t outleft, outsize, length;
  iconv_t cd;
  char *out, *outptr;
  ICONV_CONST char *input = (ICONV_CONST char *) string;

  if (!string)
    return NULL;

  length = strlen(string);

  if ((cd = iconv_open("ISO-8859-1", "UTF-8")) == (iconv_t)-1) {
    printf("convert_string(): Conversion not supported.\n");
    return NULL;
  }

  /* Due to a GLIBC bug, round outbuf_size up to a multiple of 4 */
  /* + 1 for nul in case len == 1 */
  outsize = ((length + 3) & ~3) + 1;
  out = malloc(outsize);
  outleft = outsize - 1;
  outptr = out;

 retry:
  if (iconv(cd, &input, &length, &outptr, &outleft) == -1)
  {
    int used;
    switch (errno) {
      case E2BIG:
        used = outptr - out;
        outsize = (outsize - 1) * 2 + 1;
        out = realloc(out, outsize);
        outptr = out + used;
        outleft = outsize - 1 - used;
        goto retry;
      case EINVAL:
        break;
      case EILSEQ:
        /* Invalid sequence, try to get the rest of the string */
        input++;
        length = strlen(input);
        goto retry;
      default:
        printf("Conversion failed. Inputstring: %s; Error: %s\n", string, strerror(errno));
        break;
    }
  }
  *outptr = '\0';

  iconv_close(cd);
  return out;
}
#else
static char* convert_string(const char *string)
{
  char *tempBuf = NULL;

  if (!(tempBuf = (char *) malloc(strlen(string) + 1))) {
    return NULL;
  }
  strcpy(tempBuf, string);

  return tempBuf;
}
#endif

char *get_string(FILE *fh, u_int32_t size)
{
  char *tempBuf = NULL;
  char *strBuf = NULL;
  char *cr = NULL;

  if (!(tempBuf = (char *) malloc(size + 1))) {
    return NULL;
  }
  if ((fread(tempBuf, 1, size, fh)) < size) {
    free(tempBuf);
    return NULL;
  }
  tempBuf[size] = 0; /* Null-terminate the string. */

  strBuf = convert_string(tempBuf);
  free(tempBuf);

  /* Convert carriage returns to spaces. Bit of a hack, but simpler
     than trying to figure out how to delete them from the middle
     of strings. */
  while ((cr = (char *) strchr(strBuf, 0x0D)) != NULL) {
    *cr = ' ';
  }
  return strBuf;
}

int get_ilst(FILE *fh, udta_t *udta, atom_t *atom)
{
  atom_t sub_atom, quark;
  u_int8_t Buf[4];
  int genre_max = (sizeof(genre_list) / sizeof(*genre_list));

  /* Now we can finally, *finally*, start loading up actual useful
     meta-data tags and stuffing them into the udta atom. Two levels
     of atom nesting?! Three, if you count each information atom's
     data sub-atom?! WTF, Apple? */
  do {
    atom_head(fh, &sub_atom);

#ifdef DEBUG
    printf("Got atom %s\n", sub_atom.name);
    printf("Current position in atom %s: %u\n", atom->name, atom->pos);
#endif
    if ((atomid(sub_atom.name) == atomid("\xa9""nam")) &&
        (sub_atom.size > 8)) {
      atom_head(fh, &quark); /* Get buffer length, skip data atom name */
      quark.size -= 16; /* Subtract header size, version/flags, language */
      /* Skip version/flags and language. Doing this manually, 'cause we
         can't advance the sub_atom counter until the whole atom's been
         processed. */
      fseek(fh, 8, SEEK_CUR);
      udta->title = get_string(fh, quark.size);
    } else if ((atomid(sub_atom.name) == atomid("\xA9""alb")) &&
               (sub_atom.size > 8)) {
#ifdef DEBUG
      printf("Got album atom.\n");
#endif
      atom_head(fh, &quark); /* Get buffer length, skip data atom name */
      quark.size -= 16; /* Subtract header size, version/flags, language */
      /* Skip version/flags and language. Doing this manually, 'cause we
         can't advance the sub_atom counter until the whole atom's been
         processed. */
      fseek(fh, 8, SEEK_CUR);
      udta->album = get_string(fh, quark.size);
#ifdef DEBUG
      printf("Album: %s\n", udta->album);
#endif
    } else if ((atomid(sub_atom.name) == atomid("\xA9""ART")) &&
               (sub_atom.size > 8)) {
      atom_head(fh, &quark); /* Get buffer length, skip data atom name */
      quark.size -= 16; /* Subtract header size, version/flags, language */
      /* Skip version/flags and language. Doing this manually, 'cause we
         can't advance the sub_atom counter until the whole atom's been
         processed. */
      fseek(fh, 8, SEEK_CUR);
      udta->artist = get_string(fh, quark.size);
    } else if ((atomid(sub_atom.name) == atomid("\xA9""wrt")) &&
               (sub_atom.size > 8)) {
      atom_head(fh, &quark); /* Get buffer length, skip data atom name */
      quark.size -= 16; /* Subtract header size, version/flags, language */
      /* Skip version/flags and language. Doing this manually, 'cause we
         can't advance the sub_atom counter until the whole atom's been
         processed. */
      fseek(fh, 8, SEEK_CUR);
      udta->composer = get_string(fh, quark.size);
    } else if ((atomid(sub_atom.name) == atomid("\xA9""day")) &&
               (sub_atom.size > 8)) {
      atom_head(fh, &quark); /* Get buffer length, skip data atom name */
      quark.size -= 16; /* Subtract header size, version/flags, language */
      /* Skip version/flags and language. Doing this manually, 'cause we
         can't advance the sub_atom counter until the whole atom's been
         processed. */
      fseek(fh, 8, SEEK_CUR);
      udta->date = get_string(fh, quark.size);
    } else if ((atomid(sub_atom.name) == atomid("gnre")) &&
               (sub_atom.size > 8)) {
      u_int16_t g_num;

      atom_head(fh, &quark); /* Get buffer length, skip data atom name */
      quark.size -= 16; /* Subtract header size, version/flags, language */
      /* Skip version/flags and language. Doing this manually, 'cause we
         can't advance the sub_atom counter until the whole atom's been
         processed. */
      fseek(fh, 8, SEEK_CUR);
      if ((fread(Buf, 1, 2, fh)) < 2) {
        return 0;
      }
      g_num = str2int(Buf, 2);
      g_num--; /* Genre number is 1-indexed. Decrement for array idx. */
      if (g_num > genre_max) {
        return 0;
      }
      if (!udta->genre) { /* Already have one set here by some other
                             genre tag, like (c)gen. Why there might be
                             both, I don't know. But just in case... */
        free(udta->genre);
        udta->genre = NULL;
      }
      if (!(udta->genre = (char *) malloc(strlen(genre_list[g_num]) + 1))) {
        udta->genre = NULL;
        return 0;
      }
      strcpy(udta->genre, genre_list[g_num]);
    } else if ((atomid(sub_atom.name) == atomid("\xA9""gen")) &&
               (sub_atom.size > 8)) {
      atom_head(fh, &quark); /* Get buffer length, skip data atom name */
      quark.size -= 16; /* Subtract header size, version/flags, language */
      /* Skip version/flags and language. Doing this manually, 'cause we
         can't advance the sub_atom counter until the whole atom's been
         processed. */
      fseek(fh, 8, SEEK_CUR);
      if (!udta->genre) { /* Already have one set here by some other
                             genre tag, like gnre. Why there might be
                             both, I don't know. But just in case... */
        free(udta->genre);
        udta->genre = NULL;
      }
      udta->genre = get_string(fh, quark.size);
    } else if ((atomid(sub_atom.name) == atomid("trkn")) &&
               (sub_atom.size > 8)) {
      atom_head(fh, &quark); /* Get buffer length, skip data atom name */
      ff_atom(fh, &quark, 8); /* Skip version/flags, language */
      ff_atom(fh, &quark, 2); /* Skip additional two mystery bytes */
      if ((fread(Buf, 1, 4, fh)) < 4) {
        udta->tracknum = 0;
        udta->tracks = 0;
        return 0;
      }
      quark.pos += 4;
      udta->tracknum = str2int(Buf, 2);
      udta->tracks = str2int(Buf + 2, 2);
      skip_atom(fh, &quark); /* Skip over other mystery bits */
    } else if ((atomid(sub_atom.name) == atomid("cpil")) &&
               (sub_atom.size > 8)) {
      atom_head(fh, &quark); /* Get buffer length, skip data atom name */
      ff_atom(fh, &quark, 8); /* Skip version/flags, language */
      if ((fread(Buf, 1, 1, fh)) < 1) {
        udta->is_comp = 0;
        return 0;
      }
      quark.pos++;
      udta->is_comp = Buf[0];
      skip_atom(fh, &quark); /* Skip over other mystery bits */
    } else if ((atomid(sub_atom.name) == atomid("disk")) &&
               (sub_atom.size > 8)) {
      atom_head(fh, &quark); /* Get buffer length, skip data atom name */
      ff_atom(fh, &quark, 8); /* Skip version/flags, language */
      ff_atom(fh, &quark, 2); /* Skip additional two mystery bytes */
      if ((fread(Buf, 1, 4, fh)) < 4) {
        udta->discnum = 0;
        udta->discs = 0;
        return 0;
      }
      quark.pos += 4;
      udta->discnum = str2int(Buf, 2);
      udta->discs = str2int(Buf + 2, 2);
      skip_atom(fh, &quark); /* Skip over other mystery bits */
    } else if ((atomid(sub_atom.name) == atomid("\xa9""cmt")) &&
               (sub_atom.size > 8)) {
      atom_head(fh, &quark); /* Get buffer length, skip data atom name */
      quark.size -= 16; /* Subtract header size, version/flags, language */
      /* Skip version/flags and language. Doing this manually, 'cause we
         can't advance the sub_atom counter until the whole atom's been
         processed. */
      fseek(fh, 8, SEEK_CUR);
      udta->comment = get_string(fh, quark.size);
    } else {
      skip_atom(fh, &sub_atom);
    }
    atom->pos += sub_atom.size;
#ifdef DEBUG
    printf("Current position in atom %s: %u\n", atom->name, atom->pos);
#endif
  } while (atom->pos < atom->size);

  skip_atom(fh, atom);
  return 1;
}

int get_meta(FILE *fh, udta_t *udta, atom_t *atom)
{
  atom_t child_atom;

  ff_atom(fh, atom, 4); /* Skip version/flags */
  do {
    atom_head(fh, &child_atom);

#ifdef DEBUG
    printf("Current position in atom %s: %u\n", atom->name, atom->pos);
#endif
    /* Only care about the ilst atom containing iTunes meta-data */
    if (atomid(child_atom.name) == atomid("ilst")) {
      get_ilst(fh, udta, &child_atom);
    } else {
      skip_atom(fh, &child_atom);
    }
    atom->pos += child_atom.size;
#ifdef DEBUG
    printf("Current position in atom %s: %u\n", atom->name, atom->pos);
#endif
  } while (atom->pos < atom->size);

  skip_atom(fh, atom);
  return 1;
}

int get_udta(FILE *fh, udta_t *udta, atom_t *atom)
{
  atom_t child_atom;

  do {
    /* Account for stupid QuickTime 1.0 bug in udta atom structure
       which causes it to end the atom with 4 zeroes. */
    if ((atom->size - atom->pos) < 8) { /* Not enough bytes for atom header */
      break; /* So get out of the loop, let the skip at the end clean up */
    }
    /* Okay. So we have enough bytes left for a header. Read it. */
    atom_head(fh, &child_atom);

    /* Only one atom we care about here, and that's got meta-data. */
    if (atomid(child_atom.name) == atomid("meta")) {
      get_meta(fh, udta, &child_atom);
    } else {
      skip_atom(fh, &child_atom);
    }
    atom->pos += child_atom.size;
  } while (atom->pos < atom->size);

  skip_atom(fh, atom);
  return 1;
}

void init_udta(udta_t *udta)
{
  udta->title = NULL;
  udta->album = NULL;
  udta->artist = NULL;
  udta->composer = NULL;
  udta->date = NULL;
  udta->genre = NULL;

  udta->tracknum = 0;
  udta->tracks = 0;
  udta->is_comp = 0;
  udta->discnum = 0;
  udta->discs = 0;

  udta->comment = NULL;
}

void clear_udta(udta_t *udta)
{
  if (udta->title) free(udta->title);
  if (udta->album) free(udta->album);
  if (udta->artist) free(udta->artist);
  if (udta->composer) free(udta->composer);
  if (udta->date) free(udta->date);
  if (udta->genre) free(udta->genre);

  if (udta->comment) free(udta->comment);

  init_udta(udta);
}
